# jajucha2/lidar


import requests
import base64
import subprocess
import numpy as np
import socket
import struct
import pickle
import re
import numpy as np
import cv2

SERVER_IP = "127.0.0.1"  # 서버 IP 주소
PORT = 12345            # 서버 포트 번호


def get_lidar():
    try:
        client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        
        # 서버에 연결
        client_socket.connect((SERVER_IP, PORT))

            # 사용자 입력으로 요청 생성
        request = 'request';
        if request == "exit":
            print("Exiting...")
        
        if request == "request":
            # 서버에 요청 전송
            client_socket.sendall(request.encode('utf-8'))

            # 서버로부터 데이터 수신
            data = client_socket.recv(30000)  # 한 번에 최대 4096 바이트 수신
            if not data:
                print("Server disconnected.")

            # Update the regular expression pattern to capture 'Q' as well
            pattern = r'theta:\s*([0-9.]+)\s*Dist:\s*([0-9.]+)\s*Q:\s*(\d+)'

            # Use re.findall to extract all matches at once
            matches = re.findall(pattern, data.decode('utf-8'))

            # Filter out matches where Q is '0'
            filtered_matches = [match for match in matches if match[2] != '0']

            # Convert the filtered matches to NumPy arrays
            theta_array = np.array([float(match[0]) for match in filtered_matches])
            dist_array = np.array([float(match[1]) for match in filtered_matches])

            return theta_array, dist_array
        else:
            print("Invalid input. Type 'request' to fetch data or 'exit' to quit.")

    except KeyboardInterrupt:
        print("\nClient stopped.")

    except Exception as e:
        print(f"An error occurred: {e}")

    finally:
        # 소켓 닫기
        client_socket.close()

def show_lidar(theta_array,dist_array,max_dist = 5000):
    #convert lidar data to image and send it to server

    #theta_array, dist_array = get_lidar()

        # Convert theta from degrees to radians
    #theta_rad = np.deg2rad(theta_array)
    theta_rad = np.deg2rad(-theta_array)

    # Compute Cartesian coordinates
    x = dist_array * np.cos(theta_rad)
    y = dist_array * np.sin(theta_rad)

    # Since OpenCV's origin is at the top-left corner, and y increases downwards,
    # we need to adjust the coordinates. Also, we need to scale the coordinates
    # to fit within the image size.

    # Define image size
    img_size = 800  # Adjust the image size as needed
    #max_dist = dist_array.max()
    #max_dist= 12000

    # Avoid division by zero if max_dist is zero
    if max_dist == 0:
        max_dist = 1

    # Scale factor to fit the points within the image
    scale = (img_size / 2) / max_dist

    # Shift and scale the coordinates to the center of the image
    x_img = (x * scale + img_size / 2).astype(np.int32)
    # y_img = (y * scale + img_size / 2).astype(np.int32)
    y_img = ((-y) * scale + img_size / 2).astype(np.int32)


    # Create a blank image
    image = np.zeros((img_size, img_size, 3), dtype=np.uint8)

    # Ensure the coordinates are within image boundaries
    valid_indices = (x_img >= 0) & (x_img < img_size) & (y_img >= 0) & (y_img < img_size)
    x_img = x_img[valid_indices]
    y_img = y_img[valid_indices]

    # Draw the points
    for xi, yi in zip(x_img, y_img):
        cv2.circle(image, (xi, yi), radius=2, color=(0, 255, 0), thickness=-1)

        encode_param = [int(cv2.IMWRITE_JPEG_QUALITY),90]

    image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    _, buffer = cv2.imencode('.jpg', image, encode_param)
    
    jpg_as_text = base64.b64encode(buffer).decode('utf-8')

    data = {'image': jpg_as_text}
    #data = {window_name: jpg_as_text}

    url = 'http://121.184.63.113:4000/'+ 'lidar'
    
    response = requests.post(url, json=data)
    if response.status_code != 200:
        print('Failed to send data')




from . import control,camera,lidar

# 동적 로드 관리
import importlib

def load_ai():
    """
    동적으로 ai 모듈을 로드합니다.
    """
    if "ai" not in globals():
        global ai
        ai = importlib.import_module(f"{__name__}.ai")
        print("jajucha ai loaded!")
    return ai


import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as transforms
from PIL import Image as Img
from torch.utils.data import Dataset
import os
import cv2
import numpy as np
from torchsummary import summary

from IPython.display import display, Image
import ipywidgets as widgets
from IPython.display import clear_output
import mysql.connector
from ultralytics import YOLO
import logging

logging.getLogger("ultralytics").setLevel(logging.CRITICAL)

conn = mysql.connector.connect(
    host="127.0.0.1",
    user="root",
    password="1111",
    database="datasets"
)

transform = transforms.Compose([
    transforms.Resize(224),         
    transforms.CenterCrop((224, 224)), 
    transforms.ToTensor(),
])

cursor = conn.cursor()
home_dir = os.path.expanduser("~")
device = torch.device('cuda')

def get_classes(project):
    query = "select Name from Classes where project= %s;"
    cursor.execute(query, (project,))
    rows = cursor.fetchall()
    return rows

def get_model_dir(name):
    # Use parameterized query to prevent SQL injection
    query = "SELECT model FROM Projects WHERE Name = %s"
    cursor.execute(query, (name,))
    rows = cursor.fetchall()

    # Check if any rows were returned
    if rows:
        return rows[0][0]  # Return the first element of the first row
    else:
        return None  # Return None if no rows were found
    

def show_models():
    query = "SELECT Name,Model FROM UserModels"
    cursor.execute(query)
    rows = cursor.fetchall()
    for row in rows:
        print(row)

#yolo should be added
def get_model(name):
    query = "SELECT uuid, Model FROM UserModels WHERE Name = %s"
    cursor.execute(query, (name,))
    rows = cursor.fetchone()  # Use fetchone if you're expecting only one result
    uuid, model_name = rows[0], rows[1]

    modeldir = home_dir + "/server/usermodels/" + uuid + ".pt"

    loaded_model = None
    if model_name in ['RESNET18', 'RESNET34' , 'RESNET50', 'RESNET101', 'RESNET152','EMPTY']:
        loaded_model = torch.jit.load(modeldir)
        loaded_model.eval()  # Set model to evaluation mode
    elif model_name in ['YOLO8N', 'YOLO9N', 'YOLO10N', 'YOLO11N']:
        loaded_model = YOLO(modeldir)
    return loaded_model

# Cache for model UUIDs, shapes, and loaded models
model_cache = {}

def get_model_output(input, name):
    if name not in model_cache:
        # Fetch metadata from the database if not in cache
        query = "SELECT uuid, Model FROM UserModels WHERE Name = %s"
        cursor.execute(query, (name,))
        rows = cursor.fetchone()  # Use fetchone if you're expecting only one result
        uuid, model_name = rows[0], rows[1]
        
        modeldir = home_dir + "/server/usermodels/" + uuid + ".pt"
        
        # Fetch the required output shape
        query = "SELECT shape FROM Models WHERE Name = %s"
        cursor.execute(query, (model_name,))
        shape_row = cursor.fetchone()
        shape = shape_row[0]
        
        # Load the model and cache it
        loaded_model = None
        if model_name in ['RESNET18','RESNET34' , 'RESNET50', 'RESNET101', 'RESNET152','EMPTY']:
            loaded_model = torch.jit.load(modeldir)
            loaded_model.eval()  # Set model to evaluation mode
        elif model_name in ['YOLO8N', 'YOLO9N', 'YOLO10N', 'YOLO11N']:
            loaded_model = YOLO(modeldir)
        
        # Cache model data, including the loaded model
        model_cache[name] = {
            'modeldir': modeldir,
            'model_name': model_name,
            'shape': shape,
            'model': loaded_model  # Cache the loaded model
        }
    else:
        # Fetch from cache
        modeldir = model_cache[name]['modeldir']
        model_name = model_cache[name]['model_name']
        shape = model_cache[name]['shape']
        loaded_model = model_cache[name]['model']  # Use cached model

    # Use the cached model to get the output
    if model_name in ['RESNET18','RESNET34' , 'RESNET50', 'RESNET101', 'RESNET152','EMPTY']:
        image = Img.fromarray(cv2.cvtColor(input, cv2.COLOR_BGR2RGB))
        image = transform(image)
        image = image.unsqueeze(0).to(device)
        output = loaded_model(image)
        return output
    
    elif model_name in ['YOLO8N', 'YOLO9N', 'YOLO10N','YOLO11N']:
        results = loaded_model(input)
        return results
 

def get_model_input(name):
    query = "SELECT Model FROM UserModels WHERE Name = %s"
    cursor.execute(query, (name,))
    rows = cursor.fetchall()
    model = rows[0][0]

    query = "SELECT shape FROM Models WHERE Name = %s"
    cursor.execute(query, (model,))
    rows = cursor.fetchall()
    shape = rows[0][0]
    print(shape)


# jajucha/control.py
import socket
import time
import keyboard
import re
# Server address and port
SERVER_ADDRESS = '127.0.0.1'
SERVER_PORT = 22222
UDP_PORT = 33333

last_motor_call_time = 0

def init_motor():
    send_message('45455009')

def stop_motor():
    send_message('45454500')

def set_motor(left = 0,right = 0,speed = 0):

    global last_motor_call_time
    
    # Enforce a 30Hz limit (minimum interval: 1/30 seconds = 0.0333 seconds)
    current_time = time.time()
    if current_time - last_motor_call_time < 1 / 30:  # 30Hz limit
        return  # Skip execution if called too soon

    last_motor_call_time = current_time  # Update the last call time

    #clip the steering
    if(left > 10):
        left = 10
    elif(left < -10):
        left = -10
    
    if(right > 10):
        right = 10  
    elif(right < -10):  
        right = -10

    #clip the speed
    if(speed > 30):
        speed = 30
    elif(speed < -30):
        speed = -30

    left = 45 - left
    right = 45 - right
    speed = 50 - speed 

    message = f"{left}{right}{speed}09"
    send_message(str(message))

def get_voltage():

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((SERVER_ADDRESS, UDP_PORT))
    extracted_numbers = []

    data, addr = sock.recvfrom(1024)
    output = re.findall(r'\d+', data.decode())

    voltage = (int(output[0][:3])/10 + 0.6)  # First 3 characters
    current = int(output[0][3:5])/10 # Next 2 characters
    angle = int(int(output[0][5:7])*(100/99)*3.6) # Last 2 characters

    sock.close()

    return output

def get_batt_percentage():
    voltage = get_voltage()
    batt_percentage = ((voltage/2 - 6.5) / (8.4 - 6.5) * 100)
    return batt_percentage

def get_current():
    extracted_numbers = []

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((SERVER_ADDRESS, UDP_PORT))

    data, addr = sock.recvfrom(1024)
    output = re.findall(r'\d+', data.decode())

    voltage = int(output[0][:3])/10  # First 3 characters
    current = int(output[0][3:5])/10 # Next 2 characters
    angle = int(int(output[0][5:7])*(100/99)*3.6) # Last 2 characters

    sock.close()

    return current

def get_angle():
    extracted_numbers = []

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((SERVER_ADDRESS, UDP_PORT))

    data, addr = sock.recvfrom(1024)
    output = re.findall(r'\d+', data.decode())

    voltage = int(output[0][:3])/10  # First 3 characters
    current = int(output[0][3:5])/10 # Next 2 characters
    angle = (int(output[0][5:7])*(100/99)*3.6) # Last 2 characters

    sock.close()

    return angle

def send_message(message):
    # Create a UDP socket
    client_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    try:
        # Send message to the server
        client_socket.sendto(message.encode(), (SERVER_ADDRESS, SERVER_PORT))
    finally:
        # Close the socket
        client_socket.close()
